/*
 * By: Thomas Buck <taucher.bodensee@gmail.com>
 * Visit: www.xythobuz.org
 */

#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#define SPLASHMESSAGE "SerialDebug"
#define BUFFER_SIZE 14

volatile uint8_t matrixData[5];
volatile uint16_t matrixWasShown = 0; // Increments 1562.5 times / sec => every 0,64ms
volatile uint8_t isrCounter = 0;

volatile char rxBuffer[BUFFER_SIZE];
volatile uint8_t txBuffer[BUFFER_SIZE];
volatile uint8_t rxRead = 0;
volatile uint8_t rxWrite = 0;
volatile uint8_t txRead = 0;
volatile uint8_t txWrite = 0;

unsigned char font[94][5];

void init(void);
void setVertical(uint8_t data);
void selectHorizontal(uint8_t column);
void setMatrixData(uint8_t *data, uint8_t offset);
void showForTime(uint8_t *data, uint8_t offset, uint16_t millis);
void getFont(uint8_t n, uint8_t *target, uint8_t offset);
void showSplash(void);
void moveLeft(uint8_t *data, uint8_t shiftFlag, uint8_t width);
uint8_t isZero(uint8_t *data, uint8_t offset);
void serialOpen(void);
uint8_t serialHasChar(void);
void putChar(char c);
void putString(char *data);
char getc(void);

int main(void) {
	uint8_t data[5 * 2];
	char c;
	uint8_t second = 0;

	init();
	serialOpen();
	sei(); // Enable interrupts

	showSplash();
	
	rxBuffer[rxWrite] = ' ';
	if (rxWrite < (BUFFER_SIZE - 1)) {
		rxWrite++;
	} else {
		rxWrite = 0;
	}
	
	while(1) {
		if (serialHasChar()) {
			c = getc();
			putChar(c);
			if (second == 0) {
				getFont(c, data, 0);
				second = 1;
			} else {
				getFont(c, data, 5);
				for (c = 0; c < 5; c++) {
					showForTime(data, 0, 50);
					moveLeft(data, 0, 10);
				}
			}
			setMatrixData(data, 0);
		}
	}
}

void init(void) {
	DDRB = 0xFF; // All output, matrix columns
	DDRD = 0xFF; // All output, matrix rows

	TCCR0B |=  (1 << CS01); // Prescaler: 8
	// TCCR0B |= (1 << CS00) | (1 << CS02); // Prescaler: 1024
	TIMSK |= (1 << TOIE0); // Enable overflow interrupt
}

ISR(TIMER0_OVF_vect) {
	selectHorizontal(0);
	setVertical(matrixData[isrCounter]);
	selectHorizontal(++isrCounter);
	if (isrCounter >= 5) {
		isrCounter = 0;
		matrixWasShown++;
	}
}

void setVertical(uint8_t data) {
	PORTB = data;
}

void selectHorizontal(uint8_t column) {
	PORTD = 0;
	switch(column) {
		case 1:
			PORTD |= (1 << PD6);
			break;
		case 2:
			PORTD |= (1 << PD2);
			break;
		case 3:
			PORTD |= (1 << PD3);
			break;
		case 4:
			PORTD |= (1 << PD4);
			break;
		case 5:
			PORTD |= (1 << PD5);
			break;
	}
}

void setMatrixData(uint8_t *data, uint8_t offset) {
	uint8_t i;
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		for (i = 0; i < 5; i++) {
			matrixData[i] = data[i + offset];
		}
		matrixWasShown = 0;
		isrCounter = 0;
	}
}

void showForTime(uint8_t *data, uint8_t offset, uint16_t millis) {
	// 6,4 screen refreshes per 10 milliseconds
	uint16_t roundsPassed = 0;
	millis /= 10;
	if (millis == 0) {
		millis = 1;
	}

	setMatrixData(data, offset);
	while (roundsPassed < millis) {
		if (matrixWasShown >= 6) {
			matrixWasShown = 0;
			roundsPassed++;
		}
	}
}

void getFont(uint8_t n, uint8_t *target, uint8_t offset) {
	uint8_t i;
	if ((n >= 0x21) && (n <= 0x7E)) {
		for (i = 0; i < 5; i++) {
			target[i + offset] = pgm_read_byte(&(font[n - 0x21][i]));
		}
	} else {
		for (i = 0; i < 5; i++) {
			target[i + offset] = 0;
		}
	}
	if ((isZero(target, offset)) && (n != ' ')) {
		getFont('?', target, offset);
	}
}

void showSplash(void) {
	char msg[] = SPLASHMESSAGE;
	uint8_t data[5 * 2];
	uint8_t i;
	uint8_t m;
	
	getFont(msg[0], data, 0);
	showForTime(data, 0, 400);
	for (i = 1; i < (sizeof(msg) - 1); i++) {
		getFont(msg[i], data, 5);
		for (m = 0; m < 5; m++) {
			showForTime(data, 0, 150);
			moveLeft(data, 0, 10);
		}
		showForTime(data, 0, 400);
	}
	showForTime(data, 0, 500);
}

// shiftFlag = 0  => fill up with nothing
// shiftFlag != 0 => fill up with data shifted out
// width: 5 for one screen
void moveLeft(uint8_t *data, uint8_t shiftFlag, uint8_t width) {
	uint8_t i, temp, first = data[0];
	for (i = 0; i < (width - 1); i++) {
		temp = data[i];
		data[i] = data[i + 1];
		if (i > 0) {
			data[i - 1] = temp;
		}
	}
	if (shiftFlag == 0) {
		data[(width - 1)] = 0;
	} else {
		data[(width - 1)] = first;
	}
}

uint8_t isZero(uint8_t *data, uint8_t offset) {
	uint8_t i;
	for (i = 0; i < 5; i++) {
		if (data[i + offset] != 0) {
			return 0;
		}
	}
	return 1;
}

ISR(USART_RX_vect) { // Receive complete
    rxBuffer[rxWrite] = UDR;
	if (rxWrite < (BUFFER_SIZE - 1)) {
		rxWrite++;
	} else {
		rxWrite = 0;
	}
}

ISR(USART_UDRE_vect) { // Data register empty
    if (txRead != txWrite) {
		UDR = txBuffer[txRead];
		if (txRead < (BUFFER_SIZE - 1)) {
			txRead++;
		} else {
			txRead = 0;
		}
		if (txRead == txWrite) {
			UCSRB &= ~(1 << UDRIE);
		}
	}
}

void serialOpen(void) {
	UCSRB = (1 << RXCIE) | (1 << RXEN) | (1 << TXEN); // Enable Reciever and its interrupt
	UCSRC = (1 << UCSZ1) | (1 << UCSZ0); // 8 bit chars
	UBRRH = 0;
	UBRRL = 51; // 19200 Baud
}

uint8_t serialHasChar(void) {
	if (rxRead != rxWrite) {
		return 1;
	} else {
		return 0;
	}
}

void putChar(char c) {
	while (((txWrite + 1) == txRead) || ((txRead == 0) && ((txWrite + 1) == BUFFER_SIZE))); // Buffer is full, wait!
    txBuffer[txWrite] = c;
	if (txWrite < (BUFFER_SIZE - 1)) {
		txWrite++;
	} else {
		txWrite = 0;
	}
	UCSRB |= (1 << UDRIE); // Enable Interrupt
	UCSRA |= (1 << UDRE);
}

void putString(char *data) {
	while (*data != '\0') {
		putChar(*data++);
	}
}

char getc(void) {
	char c;
	if (rxRead != rxWrite) {
		c = rxBuffer[rxRead];
		rxBuffer[rxRead] = 0;
		if (rxRead < (BUFFER_SIZE - 1)) {
			rxRead++;
		} else {
			rxRead = 0;
		}
		return c;
	} else {
		return 0;
	}
}

unsigned char font[94][5] PROGMEM = {
{0x00,0x00,0x2F,0x00,0x00},	// 0x21
{0x00,0x03,0x00,0x03,0x00},	// 0x22
{0x34,0x1C,0x36,0x1C,0x16},	// 0x23
{0x00,0x26,0x7F,0x32,0x00},	// 0x24
{0x32,0x0D,0x1E,0x2C,0x13},	// 0x25
{0x18,0x26,0x2D,0x12,0x28},	// 0x26
{0x00,0x00,0x03,0x00,0x00},	// 0x27
{0x00,0x1C,0x22,0x41,0x41},	// 0x28
{0x41,0x41,0x22,0x1C,0x00},	// 0x29
{0x00,0x0A,0x05,0x0A,0x00},	// 0x2A
{0x00,0x10,0x38,0x10,0x00},	// 0x2B
{0x00,0x80,0x60,0x00,0x00},	// 0x2C
{0x00,0x08,0x08,0x08,0x00},	// 0x2D
{0x00,0x00,0x20,0x00,0x00},	// 0x2E
{0x00,0x60,0x18,0x06,0x01},	// 0x2F
{0x00,0x1E,0x21,0x21,0x1E},	// 0x30
{0x00,0x22,0x3F,0x20,0x00},	// 0x31
{0x00,0x31,0x29,0x26,0x00},	// 0x32
{0x00,0x25,0x25,0x1A,0x00},	// 0x33
{0x00,0x0C,0x0A,0x3F,0x08},	// 0x34
{0x00,0x27,0x25,0x19,0x00},	// 0x35
{0x00,0x1E,0x25,0x25,0x18},	// 0x36
{0x00,0x01,0x39,0x05,0x03},	// 0x37
{0x00,0x1A,0x25,0x25,0x1A},	// 0x38
{0x00,0x06,0x29,0x29,0x1E},	// 0x39
{0x00,0x00,0x24,0x00,0x00},	// 0x3A
{0x00,0x80,0x64,0x00,0x00},	// 0x3B
{0x00,0x08,0x08,0x14,0x22},	// 0x3C
{0x00,0x14,0x14,0x14,0x14},	// 0x3D
{0x00,0x22,0x14,0x08,0x08},	// 0x3E
{0x00,0x01,0x29,0x05,0x02},	// 0x3F
{0x3C,0x42,0x59,0x55,0x5E},	// 0x40
{0x30,0x1C,0x12,0x1C,0x30},	// 0x41
{0x00,0x3E,0x2A,0x36,0x00},	// 0x42
{0x00,0x1C,0x22,0x22,0x22},	// 0x43
{0x00,0x3E,0x22,0x22,0x1C},	// 0x44
{0x00,0x3E,0x2A,0x2A,0x00},	// 0x45
{0x00,0x3E,0x0A,0x0A,0x00},	// 0x46
{0x00,0x1C,0x22,0x2A,0x3A},	// 0x47
{0x00,0x3E,0x08,0x08,0x3E},	// 0x48
{0x00,0x22,0x3E,0x22,0x00},	// 0x49
{0x00,0x22,0x22,0x1E,0x00},	// 0x4A
{0x00,0x3E,0x08,0x14,0x22},	// 0x4B
{0x00,0x3E,0x20,0x20,0x20},	// 0x4C
{0x3E,0x04,0x18,0x04,0x3E},	// 0x4D
{0x00,0x3E,0x04,0x08,0x3E},	// 0x4E
{0x1C,0x22,0x22,0x22,0x1C},	// 0x4F
{0x00,0x3E,0x0A,0x0A,0x04},	// 0x50
{0x1C,0x22,0x22,0x62,0x9C},	// 0x51
{0x00,0x3E,0x0A,0x14,0x20},	// 0x52
{0x00,0x24,0x2A,0x12,0x00},	// 0x53
{0x02,0x02,0x3E,0x02,0x02},	// 0x54
{0x00,0x1E,0x20,0x20,0x1E},	// 0x55
{0x00,0x0E,0x30,0x30,0x0E},	// 0x56
{0x0E,0x30,0x0C,0x30,0x0E},	// 0x57
{0x22,0x14,0x08,0x14,0x22},	// 0x58
{0x02,0x04,0x38,0x04,0x02},	// 0x59
{0x00,0x32,0x2A,0x2A,0x26},	// 0x5A
{0x00,0x00,0x7F,0x41,0x00},	// 0x5B
{0x01,0x06,0x18,0x60,0x00},	// 0x5C
{0x00,0x41,0x7F,0x00,0x00},	// 0x5D
{0x18,0x06,0x01,0x06,0x18},	// 0x5E
{0x40,0x40,0x40,0x40,0x40},	// 0x5F
{0x00,0x01,0x02,0x00,0x00},	// 0x60
{0x00,0x34,0x34,0x38,0x20},	// 0x61
{0x00,0x3F,0x24,0x24,0x18},	// 0x62
{0x00,0x18,0x24,0x24,0x00},	// 0x63
{0x18,0x24,0x24,0x3F,0x00},	// 0x64
{0x00,0x18,0x2C,0x28,0x00},	// 0x65
{0x00,0x04,0x3E,0x05,0x05},	// 0x66
{0x00,0x58,0x54,0x54,0x3C},	// 0x67
{0x00,0x3F,0x08,0x04,0x38},	// 0x68
{0x00,0x04,0x3D,0x00,0x00},	// 0x69
{0x00,0x44,0x44,0x3D,0x00},	// 0x6A
{0x00,0x3F,0x08,0x14,0x20},	// 0x6B
{0x00,0x01,0x3F,0x00,0x00},	// 0x6C
{0x3C,0x08,0x3C,0x08,0x3C},	// 0x6D
{0x00,0x3C,0x08,0x04,0x38},	// 0x6E
{0x00,0x18,0x24,0x24,0x18},	// 0x6F
{0x00,0x7C,0x24,0x24,0x18},	// 0x70
{0x18,0x24,0x24,0x7C,0x00},	// 0x71
{0x00,0x3C,0x08,0x04,0x00},	// 0x72
{0x00,0x28,0x2C,0x14,0x00},	// 0x73
{0x00,0x04,0x1E,0x24,0x04},	// 0x74
{0x00,0x1C,0x20,0x10,0x3C},	// 0x75
{0x00,0x0C,0x30,0x30,0x0C},	// 0x76
{0x0C,0x30,0x1C,0x30,0x0C},	// 0x77
{0x00,0x24,0x18,0x18,0x24},	// 0x78
{0x40,0x4C,0x70,0x30,0x0C},	// 0x79
{0x00,0x34,0x2C,0x2C,0x00},	// 0x7A
{0x00,0x08,0x36,0x41,0x00},	// 0x7B
{0x00,0x00,0x7F,0x00,0x00},	// 0x7C
{0x00,0x41,0x36,0x08,0x00},	// 0x7D
{0x10,0x08,0x08,0x10,0x08},	// 0x7E
};

// Dont have enough flash for this...
/* uint8_t waitingCounter = 0;
void waitingScreen(void) {
	uint8_t x, y, temp;
	waitingCounter++;
	if (waitingCounter > 40) {
		waitingCounter = 1;
	}
	for (x = 0; x < 5; x++) {
		matrixData[x] = 0;
	}
	temp = waitingCounter;
	for (y = 0; y < 8; y++) {
		for (x = 0; x < 5; x++) {
			if (temp == 0) {
				break;
			}
			temp--;
			setPixel(x, y);
		}
		if (temp == 0) {
			break;
		}
	}
}
void setPixel(uint8_t x, uint8_t y) {
	matrixData[x] |= (1 << y);
} */